--[[-------------------------------------------------------------------------
	QuickTargetMacros
    Copyright (C) 2010  Morsker
	Please contact Morsker through PM on forums.wowace.com.

	Aids the user in quickly making /target macros for immediate use.

	License:
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program. If not, see <http://www.gnu.org/licenses/>.

	Note:
		This AddOn's source code is specifically designed to work with
		World of Warcraft's interpreted AddOn system.
		You have an implicit license to use this AddOn with these facilities
		since that is its designated purpose as per:
		http://www.fsf.org/licensing/licenses/gpl-faq.html#InterpreterIncompat
---------------------------------------------------------------------------]]

QTMacros = LibStub("AceAddon-3.0"):NewAddon("QTMacros", "AceConsole-3.0", "AceEvent-3.0")
local QTMacros = QTMacros
local L = LibStub("AceLocale-3.0"):GetLocale("QTMacros", true)
local db

local QTM_DB_VERSION = 2 -- This is not the release version. It's for db versioning.


-- rounds off to N significant decimal places.
-- e.g. 5.975, 2 -> 5.98
local function roundOff(val, nPlaces)
	local coef = 1
	if nPlaces then
		coef = 10^nPlaces
	end
	return floor((val*coef) + 0.5) / coef
end


local function BuildMacro(template, targets)
	local t = {}
	tinsert(t, template.header and template.header or "")
	local start = 1
	if template.firstLine and #targets >= 1 then
		tinsert(t, (gsub(template.firstLine, "$s", targets[1])))
		tinsert(t, "\n")
		start = 2
	end
	if template.line then
		for i = start,#targets do
			tinsert(t, (gsub(template.line, "$s", targets[i])))
			tinsert(t, "\n")
		end
	end
	tinsert(t, template.footer and template.footer or "")
	return table.concat(t)
end


--------------------
-- AceConfig Options
--------------------

local template_options
local options
local function giveOptions()

	local function getter(info)
		return db.templates[info[#info-1]][info[#info]]
	end

	local function blockSetter(info, val)
		if #val == 0 then
			val = nil
		elseif not strfind(val, "\n$") then
			val = val.."\n"
		end
		 db.templates[info[#info-1]][info[#info]] = val
	end

	local function lineSetter(info, val)
		 db.templates[info[#info-1]][info[#info]] = (#val > 0 and val or nil)
	end

	local templateOptions = {
		name = function(info) return info[#info] end,
		type = "group",
		get = getter,
		args = {
			header = {
				name = "Header",
				order = 1,
				type = "input",
				multiline = true,
				width = "full",
				desc = "macro header, before targetting",
				set = blockSetter,
			},
			firstLine = {
				name = "First Line",
				order = 2,
				type = "input",
				width = "full",
				desc = "use this if the first line should be different; leave blank otherwise",
				set = lineSetter,
			},
			line = {
				name = "Line",
				order = 3,
				type = "input",
				width = "full",
				desc = "template for each line of targetting,  use $s to include the name",
				set = lineSetter,
			},
			footer = {
				name = "Footer",
				order = 4,
				type = "input",
				multiline = true,
				width = "full",
				desc = "macro footer, after targetting",
				set = blockSetter,
			},
		},
	}

	options = {
		name = "Quick Target Macros",
		type = "group",
		childGroups = "tree",
		args = {
			bind = {
				name = "Bind",
				type = "execute",
				-- dialogHidden = true,
				desc = "Choose a keybind for the macro.",
				func = function(info) QTMacros:ShowKeybindFrame() end,
			},
			show = {
				name = "Show",
				type = "execute",
				dialogHidden = true,
				desc = "Show the QT button.",
				func = function(info) QTMacros:ActivateButton() end,
			},
			config = {
				name = "Config",
				type = "execute",
				dialogHidden = true,
				desc = "Open config UI.",
				func = function(info) InterfaceOptionsFrame_OpenToCategory("Quick Target Macros") end,
			},
			template = {
				name = "Templates",
				type = "group",
				args = {
					-- filled in below, one entry per each in db.templates
				},
			},
		},
	}
	for k,v in pairs(db.templates) do -- FIXME: why does pairs work, but not ipairs?
		options.args.template.args[k] = templateOptions
	end
	_G.DaveTest = options
	giveOptions = function() return options end
	return options
end


-----------------
-- addon QTMacros
-----------------


-- Create the main QT button, and assign it to self.mainButton
-- This is the button that holds the macrotext, and assigns a keybind to it when visible.
function QTMacros:InitMainButton()
	QTMacros.InitMainButton = nil
	local f = CreateFrame("Button", "QTMacrosMainButton", nil,
		"SecureHandlerShowHideTemplate,ActionButtonTemplate,SecureActionButtonTemplate")
	f.icon = _G[f:GetName().."Icon"]
	f.hotkey = _G[f:GetName().."HotKey"]
	f:SetFrameStrata("MEDIUM")
	f:SetMovable(true)
	f:SetClampedToScreen(true)
	f:Hide()

	if not db.buttonAnchor then
		db.buttonAnchor = { "CENTER", -100, -30 }
	end
	f:SetPoint(unpack(db.buttonAnchor))
	f:SetAttribute("type", "macro")
	if db.macroText then
		f:SetAttribute("macrotext", db.macroText)
	end
	f:SetAttribute("_hideme", [[ self:Hide() ]])
	f:SetAttribute("ctrl-type1", "hideme")
	f:SetAttribute("_onshow", [[
		local keybind = self:GetAttribute("keybind")
		if keybind then
			self:SetBindingClick(true, keybind, self:GetName(), "LeftButton")
		end
	]])
	f:SetAttribute("_onhide", [[
		self:ClearBindings()
	]])
	f:SetNormalTexture(nil)

	f:RegisterForClicks("AnyUp")
	f:RegisterForDrag("LeftButton")
	f.icon:SetTexture([[Interface\Addons\QuickTargetMacros\images\QT_icon.tga]])
	
	function f:ShowTooltip()
		if not InCombatLockdown() then
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
			GameTooltip:SetText("Quick Target Macros")
			GameTooltip:AddLine(self:GetAttribute("macrotext") or "")
			GameTooltip:AddLine(" ")
			GameTooltip:AddDoubleLine(L["Click"], L["Target"])
			GameTooltip:AddDoubleLine(L["Ctrl-Click"], L["Dismiss"])
			GameTooltip:AddDoubleLine(L["Alt-Click"], L["Move"])
			GameTooltip:Show();
		end
	end
	
	f:SetScript("OnEnter", function(self)
		self:ShowTooltip()
	end)
	
	f:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	
	f:SetScript("OnDragStart", function(self)
		if IsAltKeyDown() then
			GameTooltip:Hide()
			self:StartMoving()
		end
	end)
	
	f:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		self:SetUserPlaced(false)
		if self:GetNumPoints() > 0 then
			local point,_,_,x,y = self:GetPoint(1)
			db.buttonAnchor[1] = point
			db.buttonAnchor[2] = roundOff(x, 2)
			db.buttonAnchor[3] = roundOff(y, 2)
		end
		self:ShowTooltip()
	end)
	
	self.mainButton = f
end


-- You can call this multiple times. It only does init the first time.
function QTMacros:ShowKeybindFrame(showButtonAfter)
	local f = CreateFrame("Frame")
	f:SetFrameStrata("DIALOG")
	f:EnableKeyboard(true)
	f:SetClampedToScreen(true)
	f:Hide()
	f:SetPoint("CENTER", 0, UIParent:GetHeight() / 7)
	f:SetBackdrop({
		bgFile=[[Interface\Tooltips\UI-Tooltip-Background]],
		edgeFile=[[Interface\Tooltips\UI-Tooltip-Border]], tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	f:SetBackdropColor(0, 0, 0)

	f.text = f:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	f.text:SetPoint("TOP", 0, -20)
	f.text:SetText(L[ [=[
Press a key to bind the /target macro.

Or press ESC to cancel.]=] ])

	f.text2 = f:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	f.text2:SetPoint("BOTTOM", 0, 20)
	f.text2:SetText(L[ [=[
(This is a temporary "override" that only lasts as long as you're
using the macro. It won't remove your previous binding.)]=] ])

	f:SetScript("OnShow", function(self)
		local width = max(self.text:GetWidth(), self.text2:GetWidth()) + 40
		local height = self.text:GetHeight() + self.text2:GetHeight() + 60
		self:SetWidth(width)
		self:SetHeight(height)
		
		self:RegisterEvent("PLAYER_REGEN_DISABLED") -- so we can hide in combat
	end)
	
	f:SetScript("OnHide", function(self)
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	end)
	
	f:SetScript("OnEvent", function(self, event, ...)
		if event == "PLAYER_REGEN_DISABLED" then
			QTMacros:Print(L["Combat! Aborting keybind selection."])
			self:Hide() -- hide in combat
		end
	end)

	-- These generate OnKeyDown events, but aren't bindable and should be discarded.
	local unbindableKeys = {
		LALT = 1,
		LCTRL = 1,
		LSHIFT = 1,
		RALT = 1,
		RCTRL = 1,
		RSHIFT = 1,
		UNKNOWN = 1,
	}

	f:SetScript("OnKeyDown", function(self, key)
		if unbindableKeys[key] then
			return
		end
		if key == "ESCAPE" then
			self:Hide()
			return
		end
		if GetBindingFromClick(key) == "SCREENSHOT" then
			TakeScreenshot()
			return
		end
		if IsShiftKeyDown() then
			key = "SHIFT-"..key
		end
		if IsControlKeyDown() then
			key = "CTRL-"..key
		end
		if IsAltKeyDown() then
			key = "ALT-"..key
		end

		QTMacros:SetKeybind(key)
		self:Hide()
		if self.showButtonAfter then
			QTMacros:ActivateButton()
			self.showButtonAfter = nil
		end
	end)
	
	self.keybindFrame = f
	function QTMacros:ShowKeybindFrame(showButtonAfter)
		self.keybindFrame.showButtonAfter = (showButtonAfter and 1 or nil)
		self.keybindFrame:Show()
	end
	return self:ShowKeybindFrame(showButtonAfter)
end


function QTMacros:SetKeybind(key)
	self.mainButton:SetAttribute("keybind", key)
	db.keybind = key
	if key then
		self.mainButton.hotkey:SetText(GetBindingText(key, "KEY_", true))
	else
		self.mainButton.hotkey:SetText(nil)
	end
end


function QTMacros:ActivateButton()
	if self.mainButton:GetAttribute("keybind") then
		self.mainButton:Show()
	else
		self:ShowKeybindFrame(true)
	end
end


-- calling this with "/qt" creates a method self:SlashQT and calls RegisterChatCommand("qt", "SlashQT", false)
function QTMacros:BuildTemplateSlashCommand(slashTemplateName)
	if not strfind(slashTemplateName, "^/") then fail("'"..slashTemplateName.."' not a /slash command") end
	local templateName = slashTemplateName:sub(2):lower() -- "qt"
	local methodName = "Slash"..(templateName:upper()) -- "SlashQT"

	self[methodName] = function(self, input) -- function self:SlashQT(input)
		local targets = {strsplit(",", input)}
		for i = 1,#targets do
			targets[i] = strtrim(targets[i])
		end
		local macroText = BuildMacro(db.templates[slashTemplateName], targets)
		-- TODO: refactor to self:SetMacroText and make it combat-aware
		self.mainButton:SetAttribute("macrotext", macroText)
		db.macroText = macroText
		self:ActivateButton()
	end
	self:RegisterChatCommand(templateName, methodName, false)
end


function QTMacros:ReleaseSlashCommand(slashTemplateName)
	if not strfind(slashTemplateName, "^/") then fail("'"..slashTemplateName.."' not a /slash command") end
	local templateName = slashTemplateName:sub(2):lower()
	local methodName = "Slash"..(templateName:upper())

	self:UnregisterChatCommand(templateName)
	self[methodName] = nil
end


function QTMacros:BuildAllTemplateSlashCommands()
	if not self.builtSlashCommands then
		for k,v in pairs(db.templates) do
			if strfind(k, "^/") then -- template key starts with "/"; it's a slash command
				self:BuildTemplateSlashCommand(k)
			end
		end
		self.builtSlashCommands = true
	end
end


function QTMacros:OnChatCommand(input)
	LibStub("AceConfigCmd-3.0"):HandleCommand("qtmacros", "QTMacros", input)
end


function QTMacros:PLAYER_LOGOUT()
	-- db.isShown
	db.isShown = self.mainButton:IsShown()
end


function QTMacros:UpgradeDatabase()
	if db.db_version == nil then
		-- The first real version is 1. Old experimental versions had no db_version; we treat them as 0.
		db.db_version = 0
	end
	if db.db_version < 1 then
		-- This was for saving an actionID to drag a macro onto the bars in its place. Version 1 no longer uses it.
		db.savedActionID = nil
	end
	if db.db_version < 2 then
		-- first, the version 1 init logic
		if not db.cmdTemplate then
			db.cmdTemplate = {
				firstLine = "/tar $s",
				line = "/tar [noexists][dead] $s",
			}
		end
		-- now, convert version 1 to version 2
		db.templates = { ["/qt"] = db.cmdTemplate }
		db.cmdTemplate = nil
	end
	db.db_version = QTM_DB_VERSION
end


function QTMacros:OnInitialize()
	self.OnInitialize = nil
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("QTMacros", giveOptions)
	local optFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("QTMacros", "Quick Target Macros")

	self:RegisterChatCommand("quicktargetmacros", "OnChatCommand", true)
	self:RegisterChatCommand("qtmacros", "OnChatCommand", true)
	self:RegisterChatCommand("qtm", "OnChatCommand", true)

	QTMacrosPerCharDB = QTMacrosPerCharDB or {}
	db = QTMacrosPerCharDB
	if db.db_version ~= QTM_DB_VERSION then
		self:UpgradeDatabase()
	end
	self.UpgradeDatabase = nil
	if not db.templates then
		db.templates = {
			["/qt"] = {
				firstLine = "/tar $s",
				line = "/tar [noexists][dead] $s",
			}
		}
	end
	if not db.templates["QUEST"] then
		db.templates["QUEST"] = {
			firstLine = "/targetexact $s",
			line = "/targetexact [noexists][dead] $s",
		}
	end

	self:InitMainButton()

	if db.keybind then
		self:SetKeybind(db.keybind)
	end
	if db.isShown then
		self.mainButton:Show()
	end

	-- We wait until the last minute to register commands, so they supersede other addons'.
	-- This is appropriate because the users chooses these command names, and the user's
	-- choice should have priority.
	if IsLoggedIn() then
		self:BuildAllTemplateSlashCommands()
	else
		function self:PLAYER_LOGIN()
			self:UnregisterEvent("PLAYER_LOGIN")
			self.PLAYER_LOGIN = nil
			self:BuildAllTemplateSlashCommands()
		end
		self:RegisterEvent("PLAYER_LOGIN")
	end
end


function QTMacros:OnEnable()
	self:RegisterEvent("PLAYER_LOGOUT")
end


function QTMacros:OnDisable()
end
